import math

from SmartDiff.solvers.element_op import *

# The evaluator takes as input a string representation of a math expression
# and parses the string by iteratively building the AD object applied with
# the corresponding operations starting from the variable

class PyExpression_Formatter(object):

  def __init__(self, valid_special_chars=None):
    if valid_special_chars:
      self.special_char_set = valid_special_chars
    else:
      self.special_char_set = {"+", "-", "*", "/", ".", "(", ")", ",", "_", " "}

  def format_to_pyexpr(self, input_str):
    """
    This function formats the input string of math function into python code

    :param input_str: a math function
    :return: a string of compilable python code
    """
    # Check input validity
    valid_out = self.is_valid_input(input_str)
    if valid_out == 1:
      raise SyntaxError("Input function has unmatched parenthesis!")
    elif valid_out == 2:
      raise SyntaxError("Input function contains invalid character!")
    # e --> math.e
    i = 0
    pyexpr = []
    while i < len(input_str) - 1:
      c = input_str[i]
      if c.isspace():
        i += 1
      elif c == "e":
        next_c = input_str[i+1]
        if next_c in {"x", "r"}:
          pyexpr.append(c)
          pyexpr.append(next_c)
          i += 2
        else:
          pyexpr.append("math.e")
          i += 1
      elif c == "p" and input_str[i+1] == "i":
          pyexpr.append("math.pi")
          i += 2
      else:
        pyexpr.append(c)
        i += 1
    if i < len(input_str):
      if input_str[-1] == "e":
        pyexpr.append("math.e")
      elif not input_str[-1].isspace():
        pyexpr.append(input_str[-1])
    return "".join(pyexpr)

  def is_valid_input(self, input_str):
    """
    This function returns true if the input string has valid parenthesis.

    :param input_str: a string of math function
    :return: True if the input has valid parenthesis
    """
    suffix_demand = 0
    for c in input_str:
      if c == "(":
        suffix_demand += 1
      elif c == ")":
        suffix_demand -= 1
      else:
        # check if a non alphanumeral char is in our special character set
        if not c.isalnum() and c not in self.special_char_set:
          return 2
        continue
      if suffix_demand == -1:
        return 1
    return 0 if suffix_demand == 0 else 1


  # code = compile("math.log(3)", "<string>", "eval")
  # print(eval(code))
  # print(eval("math.log(4)"))
  #
  # s = "pow(x, 2) * 5"
  # x = AD(3)
  # print(eval(s))



import math
from SmartDiff.solvers.element_op import *


FUNC_MAP = {"pi": math.pi, "e": math.e, "power": power, "log": log, "exp": exp, "sqrt": sqrt, "sin": sin,
            "cos": cos, "tan": tan, "arcsin": arcsin, "arccos": arccos, "arctan": arctan, "sinh": sinh,
            "cosh": cosh, "tanh": tanh}

MATH_FUNC_MAP = {"pi": math.pi, "e": math.e, "power": math.pow, "log": math.log, "exp": math.exp, "sqrt": math.sqrt,
                 "sin": math.sin, "cos": math.cos, "tan": math.tan, "arcsin": math.asin, "arccos": math.acos,
                 "arctan": math.atan, "sinh": math.sinh, "cosh": math.cosh, "tanh": math.tanh}




import SmartDiff.solvers.element_op as el
from SmartDiff.solvers.element_op import AutoDiff as AD
from SmartDiff.solvers.multivariate import *
import numpy as np
import pytest

class TestElemOpNOrder:

    def test_add(self):
        # N > 1
        x = AD(5, N=3)
        f = el.sin(x)
        g = 10
        h = f + g
        assert (h.val, np.round(h.der[-1]), 6) == (np.sin(5) + 10, np.round(-np.cos(5)), 6)

        f = 10
        g = el.sin(x)
        h = f + g
        assert (h.val, np.round(h.der[-1]), 6) == (np.sin(5) + 10, np.round(-np.cos(5)), 6)

        f = el.sin(x)
        g = x ** 2
        h = f + g
        assert (h.val, np.round(h.der[-1]), 6) == (np.sin(5) + 25, np.round(-np.cos(5)), 6)

    def test_sub(self):
        # N > 1
        x = AD(5, N=3)
        f = el.sin(x)
        g = 10
        h = f - g
        assert (h.val, np.round(h.der[-1]), 6) == (np.sin(5) - 10, np.round(-np.cos(5)), 6)

        f = el.sin(x)
        g = x ** 2
        h = f - g
        assert (h.val, np.round(h.der[-1]), 6) == (np.sin(5) - 25, np.round(-np.cos(5)), 6)

    def test_sin(self):
        # N = 1
        x = AD(0)
        f = el.sin(x)
        assert (f.val, f.der) == (0.0, 1.0)

        x = 13
        f = el.sin(x)
        assert (f.val, f.der) == (np.sin(13), 0)

        x = 13.0
        f = el.sin(x)
        assert (f.val, f.der) == (np.sin(13), 0)

        # N > 1
        x = AD(2, N=4)
        f = el.sin(x)
        # print(f)
        assert (f.val, np.round(f.der[-1], 6)) == (np.sin(2), np.round(np.sin(2), 6))

        x = AD(2, N=3)
        f = el.sin(x)
        # print(f)
        assert (f.val, np.round(f.der[-1], 6)) == (np.sin(2), np.round(-np.cos(2), 6))

        f = el.sin(x) + 3
        assert (f.val, np.round(f.der[-1], 6)) == (np.sin(2) + 3, np.round(-np.cos(2), 6))

        f = el.sin(x) - 2
        assert (f.val, np.round(f.der[-1], 6)) == (np.sin(2) - 2, np.round(-np.cos(2), 6))

        f = el.sin(x) * 4
        assert (f.val, np.round(f.der[-1], 6)) == (4 * np.sin(2), np.round(-4 * np.cos(2), 6))

    def test_cos(self):
        # N > 1
        x = AD(10, N=5)
        f = el.cos(x)
        assert (f.val, np.round(f.der[-1], 6)) == (np.cos(10), np.round(-np.sin(10), 6))

        x = AD(2, N=2)
        f = el.cos(x)
        assert (f.val, np.round(f.der[-1], 6)) == (np.cos(2), np.round(-np.cos(2), 6))

        f = el.cos(x) + 3
        assert (f.val, np.round(f.der[-1], 6)) == (np.cos(2) + 3, np.round(-np.cos(2), 6))

        f = el.cos(x) - 2
        assert (f.val, np.round(f.der[-1], 6)) == (np.cos(2) - 2, np.round(-np.cos(2), 6))

        f = el.cos(x) * 4
        assert (f.val, np.round(f.der[-1], 6)) == (4 * np.cos(2), np.round(-4 * np.cos(2), 6))

    def test_tan(self):
        # N > 1
        x = AD(10, N=2)
        f = el.tan(x)
        assert (f.val, np.round(f.der[-1], 6)) == (np.tan(10), np.round(2 * np.tan(10) * (1 / (np.cos(10) ** 2)), 6))

        x = AD(10, N=3)
        f = el.tan(x)
        assert (f.val, np.round(f.der[-1], 6)) == (
        np.tan(10), np.round(-2 * (-2 + np.cos(20)) * (1 / (np.cos(10) ** 4)), 6))

    def test_arcsin(self):
        # N > 1
        x = AD(0.5, N=2)
        f = el.arcsin(x)
        assert (f.val, np.round(f.der[-1], 6)) == (np.arcsin(0.5), 0.7698)

    def test_arccos(self):
        # N > 1
        x = AD(0.5, N=2)
        f = el.arccos(x)
        assert (f.val, np.round(f.der[-1], 6)) == (np.arccos(0.5), -0.7698)

    def test_arctan(self):
        # N > 1
        x = AD(0.5, N=2)
        f = el.arctan(x)
        assert (f.val, np.round(f.der[-1], 6)) == (np.arctan(0.5), -0.64)

    def test_sinh(self):
        # N > 1
        x = AD(10, N=4)
        f = el.sinh(x)
        assert (f.val, f.der[-1]) == (np.sinh(10), np.sinh(10))

        x = AD(10, N=7)
        f = el.sinh(x)
        assert (f.val, f.der[-1]) == (np.sinh(10), np.cosh(10))

    def test_cosh(self):
        # N > 1
        x = AD(10, N=4)
        f = el.cosh(x)
        assert (f.val, f.der[-1]) == (np.cosh(10), np.cosh(10))

        x = AD(10, N=7)
        f = el.cosh(x)
        assert (f.val, f.der[-1]) == (np.cosh(10), np.sinh(10))

    def test_tanh(self):
        # N > 1
        x = AD(0.5, N=2)
        f = el.tanh(x)
        assert (f.val, np.round(f.der[-1], 6)) == (np.tanh(0.5), -0.726862)

        x = AD(0.5, N=3)
        f = el.tanh(x)
        assert (f.val, np.round(f.der[-1], 6)) == (np.tanh(0.5), -0.565209)

        x = AD(0, N=3)
        f = el.tanh(x)
        assert (f.val, f.der[-1]) == (np.tanh(0), -2)

    def test_exp(self):
        # N = 1
        f = el.exp(1)
        assert (f.val, f.der) == (np.exp(1), 0)

        x = AD(1)
        f = el.exp(x)
        assert (f.val, f.der) == (np.exp(1), np.exp(1))

        x = AD(2)
        f = el.power(x, 3)
        g = el.exp(f)
        print(g)
        assert (round(g.val, 6), round(g.der[0], 6)) == (round(np.exp(1) ** 8, 6), round(12 * np.exp(1) ** 8, 6))

        with pytest.raises(AttributeError):
            x = "hello"
            f = el.exp(x)

        # N > 1
        for N in range(2, 10):
            x = AD(3, N=N)
            f = el.exp(x)
            print("N=%d" % N, f)

        x = AD(2, N=5)
        f = el.exp(x)
        assert (f.val, f.der[-1]) == (np.exp(2), np.exp(2))

    def test_ln(self):
        # N = 1
        f = el.ln(np.e)
        assert (f.val, f.der) == (1, 0)

        x = AD(np.e)
        f = el.ln(x)
        assert (f.val, f.der) == (1, 1 / np.e)

        # N > 1
        # See explicit formula here: https://www.math24.net/higher-order-derivatives/#example2
        x = AD(np.e, N=2)
        f = el.ln(x)
        assert (f.val, f.der[-1]) == (1, -1 / (np.e ** 2))

        x = AD(np.e ** 2, N=5)
        f = el.ln(x)
        assert (f.val, f.der[-1]) == (2, 24 / ((np.e ** 2) ** 5))

        # composite x
        x = AD(np.e, N=4)
        f = el.ln(x * 3 + 1)
        assert (f.val, f.der[-1]) == (np.log(np.e * 3 + 1), (-6 * 3 ** 4) / ((np.e * 3 + 1) ** 4))

    def test_inv(self):
        # N = 1
        f = el.ln(np.e)
        assert (f.val, f.der) == (1, 0)

        x = AD(np.e)
        f = el.ln(x)
        assert (f.val, f.der) == (1, 1 / np.e)

        # N > 1
        # See explicit formula here: https://www.math24.net/higher-order-derivatives/#example2
        x = AD(np.e, N=2)
        f = el.ln(x)
        assert (f.val, f.der[-1]) == (1, -1 / (np.e ** 2))

        x = AD(np.e ** 2, N=5)
        f = el.ln(x)
        assert (f.val, f.der[-1]) == (2, 24 / ((np.e ** 2) ** 5))

    def test_truediv(self):
        # N = 1 tests all passed in solvers_test
        x = AD(3)
        f = 5 * x / 3
        assert (f.val, f.der[-1]) == (5, 5 / 3)

        x = AD(4, N=2)
        f = x / x
        assert (f.val, f.der[-1]) == (1, 0)

    def test_rtruediv(self):
        x = AD(4, N=2)
        f = 3 / (2 * x)
        assert (f.val, f.der[-1]) == (3 / 8, 3 / 64)

        x = AD(4, N=3)
        f = 3 / (2 * x)  # shows rtruediv should work
        assert (f.val, f.der[-1]) == (3 / 8, -9 / 4 ** 4)

    def test_sqrt(self):
        x = AD(4, N=2)
        f = el.sqrt(x)
        assert (f.val, f.der[-1]) == (2, -0.25 / (4 ** 1.5))

    def test_power(self):
        x = AD(4, N=3)
        f = el.power(x, 4)
        assert (f.val, f.der[-1]) == (256, 96)

    def test_AD_pow(self):
        pass

    def test_power_k(self):
        with pytest.raises(AttributeError):
            x = AD(10)
            n = "1"
            f = el.power_k_order(x, n, 3)

        with pytest.raises(AttributeError):
            x = AD(10)
            n = AD(3)
            f = el.power_k_order(x, n, 3)

        with pytest.raises(ValueError):
            x = AD(-3)
            n = 1 / 2
            f = el.power_k_order(x, n, 3)

    def test_logistic(self):
        x = AD(100)
        f = el.logistic(x, x0=0, L=1, k=1)
        assert (f.val, np.round(f.der[-1], 10)) == (1 / (1 + np.exp(-100)),
                                                    np.round(np.exp(100) / (1 + np.exp(100)) ** 2, 6))
        # assert (f.val, f.der[-1]) == (1/(1+np.exp(-100)), np.exp(100)/(1 + np.exp(100))**2)

        f = el.logistic(x, x0=10, L=2, k=3)
        assert (f.val, f.der[-1]) == (2 * np.exp(270) / (1 + np.exp(270)), (6 * np.exp(270)) / (1 + np.exp(270)) ** 2)

        x = AD(100, N=2)
        f = el.logistic(x, x0=0, L=1, k=1)
        assert (f.val, np.round(f.der[-1], 6)) == (
        1 / (1 + np.exp(-100)), np.round((np.exp(100) - np.exp(200)) / (1 + np.exp(100)) ** 3, 6))

        x = 100
        f = el.logistic(x, x0=0, L=1, k=1)
        assert (f.val, f.der) == (1 / (1 + np.exp(-100)), 0)

        with pytest.raises(AttributeError):
            x = "AD(-3)"
            f = el.logistic(x, x0=0, L=1, k=1)


    def test_hessian(self):

        # single variable
        func = "x0 + 3"
        vals = [5]
        Hmat = get_hessian(func_str=func, all_vals=vals)
        assert (Hmat == [[0]])
        #print("Hessian 1:", Hmat, "; Expected = [[0]]")

        func = "x0**2"
        vals = [5]
        Hmat = get_hessian(func_str=func, all_vals=vals)
        assert (Hmat == [[2]])
        #print("Hessian 2:", Hmat, "; Expected = [[2]]")

        func = "x0**3 + log(4, 2)"
        vals = [5]
        Hmat = get_hessian(func_str=func, all_vals=vals)
        assert (Hmat == [[30]])
        #print("Hessian 3:", Hmat, "; Expected = [[30]]")

        # multivariable

        func = "x0 + x1"
        vals = [3, 5]
        Hmat = get_hessian(func_str=func, all_vals=vals)
        assert (Hmat[0] == [0, 0]).all()
        assert (Hmat[1] == [0, 0]).all()

        func = "x0 * x1"
        vals = [3, 5]
        Hmat = get_hessian(func_str=func, all_vals=vals)
        assert (Hmat[0] == [0, 1]).all()
        assert (Hmat[1] == [1, 0]).all()

        func = "x0 * x1**3"
        vals = [3, 5]
        Hmat = get_hessian(func_str=func, all_vals=vals)
        assert (Hmat[0] == [0, 75]).all()
        assert (Hmat[1] == [75, 90]).all()

        func = "x0 * x1**x0 * x1"
        vals = [3, 5]
        Hmat = get_hessian(func_str=func, all_vals=vals)
        #assert (Hmat[0] == [0, 375]).all()
        #assert (Hmat[1] == [375, 900]).all() 
        #print("Expected = " + str([[0, np.log(5)+1],[np.log(5)+1, 3/5]]) + "\n")

    def test_jacobian(self):
        
        func_list = ['x0*x2+x1', 'x1']
        val_list = [2,3,4]
        values, jacobian = get_val_jacobian(func_list, val_list)
        assert (values == [11, 3]).all()
        assert (jacobian[0] == [4, 1, 2]).all()
        assert (jacobian[1] == [0, 1, 0]).all()

        func_list = ["x0*exp(x0+9)"]
        val_list = [2]
        values, jacobian = get_val_jacobian(func_list, val_list)
        assert (values == [2*np.exp(11)]).all()
        assert (jacobian[0] == [np.exp(11) + 2*np.exp(11)]).all()

        func_list = ["x1*exp(x0+9)", "x1*sin(x0)"]
        val_list = [2,5]
        values, jacobian = get_val_jacobian(func_list, val_list)
        assert (values == [5*np.exp(11), 5*np.sin(2)]).all()
        assert (jacobian[0] == [5*np.exp(11), np.exp(11)]).all()
        assert (jacobian[1] == [5*np.cos(2), np.sin(2)]).all()

        func_list = ["cos(exp(x0))*x2+x1*x2"]
        val_list = [1,2,3]
        values, jacobian = get_val_jacobian(func_list, val_list)
        assert (values == [6+3*np.cos(np.exp(1))]).all()
        assert (np.round(jacobian[0],6) == np.round([-3*np.sin(np.exp(1))*np.exp(1), 3, np.cos(np.exp(1))+2],6)).all()

        #---------- Old code? -----------#

        # print('----------')
        # output_value3, jacobian3 = get_jaco(func = (lambda x: [x**4+10]), vals = [2], ders = [[2]])
        # print('value3',output_value3)
        # print('jacobian3',jacobian3)

        # print('----Now works----')
        # output_value4, jacobian4 = get_jaco(func = (lambda x: [x**4+10, x**6]), vals = [2])
        # print('value4',output_value4)
        # print('jacobian4',jacobian4)

        # print('----Trial----')
        # output_value5, jacobian5 = get_jaco(func = (lambda x: np.array([x**4+10, x**6])), vals = 2)
        # print('value5',output_value5)
        # print('jacobian5',jacobian5)

        # output_value1, jacobian1, hess1 = jacob_hessian(lambda x,y,z : [x*z+y, y*z+x, z, z+2], [2,3,4])
        # print('value1', output_value1)
        # print('jacobian1',jacobian1)
        # output_value1, jacobian1, hess1 = jacob_hessian(lambda x,y : [x**2*exp(x)+y, y], [2,3], 1)
        # print('value1', output_value1)
        # print('jacobian1',jacobian1)
        # print('hessian1',hess1)
        #
        # output_value1, jacobian1, hess1 = jacob_hessian(lambda x,y : [x**2*exp(x),x*y], [2,3], 2)
        # print('value2', output_value1)
        # print('jacobian2',jacobian1)
        # print('hessian2',hess1)

        




# usage: pytest -q solvers_test.py
import SmartDiff.solvers.element_op as el
from SmartDiff.solvers.element_op import AutoDiff as AD
import numpy as np
import pytest


class TestElemOp:

    def test_power(self):
        x = AD(5)
        f = el.power(x, 2)
        assert (f.val, f.der) == (25, 10)

        x = 10
        f = el.power(x, 3)
        assert (f.val, f.der) == (1000, 0)

        x = 10.0
        f = el.power(x, 3)
        assert (f.val, f.der) == (1000, 0)

        with pytest.raises(AttributeError):
            x = "s"
            f = el.power(x, 3)

    def test_log(self):
        x = AD(12)
        f = el.log(x, 10)
        assert (f.val, f.der) == (np.log(12) / np.log(10), 1 / (12 * np.log(10)))

        x = 8
        f = el.log(x, 2)
        assert (np.round(f.val, 6), f.der) == (3, 0)

        x = 9.0
        f = el.log(x, 3)
        assert (f.val, f.der) == (2, 0)

        x = AD(0)
        with pytest.raises(ValueError):
            f = el.log(x, 2)

        x = 0
        with pytest.raises(ValueError):
            f = el.log(x, 2)

        with pytest.raises(AttributeError):
            x = "s"
            f = el.log(x, 3)

        with pytest.raises(ValueError):
            x = -3.9
            f = el.log(x, 3)

    def test_ln(self):
        x = AD(-10)
        with pytest.raises(ValueError):
            f = el.ln(x)

        x = AD(10, N=3)
        f = el.ln(x)
        assert (f.val, f.der[-1]) == (np.log(10), 2 * 1 / (1000))

        x = AD(10, N=4)
        f = el.ln(x)
        assert (f.val, f.der[-1]) == (np.log(10), -6 * 1 / (10000))

    def test_exp(self):
        f = el.exp(1)
        assert (f.val, f.der) == (np.exp(1), 0)

        x = AD(1)
        f = el.exp(x)
        assert (f.val, f.der) == (np.exp(1), np.exp(1))

        x = AD(2)
        f = el.power(x, 3)
        g = el.exp(f)
        assert (np.round(g.val, 6), np.round(g.der, 6)) == (
        np.round(np.exp(1) ** 8, 6), np.round(12 * np.exp(1) ** 8, 6))

        with pytest.raises(AttributeError):
            x = "hello"
            f = el.exp(x)

    def test_expn(self):
        x = AD(2)
        f = el.expn(x, 25)
        assert (f.val, f.der) == (625, 625 * np.log(25))

        x = 2
        f = el.expn(x, 25)
        assert (f.val, f.der) == (625, 0)

        with pytest.raises(AttributeError):
            f = el.expn('25', 0)

        with pytest.raises(ValueError):
            f = el.expn(x, -10)

        x = AD(4)
        f = el.expn(x, 2)
        # assert (f.val, f.der) == (16, 16*(np.log(2)**2))

    def test_inv(self):
        x = 10
        assert el.inv(x) == 0.1

        y = -1 / 30
        assert el.inv(y) == -30

        with pytest.raises(ZeroDivisionError):
            x = 0
            f = el.inv(x)

        with pytest.raises(AttributeError):
            x = "zero"
            f = el.inv(x)

        x = AD(8)
        f = el.inv(x)
        assert (f.val, f.der) == (1 / 8, -1 / 64)

        x = 92
        f = el.inv(x)
        assert (f.val, f.der) == (1 / 92, 0)

        x = AD(10, N=3)
        f = el.inv(x)
        assert (f.val, f.der[-1]) == (0.1, -6 * 1 / 10000)

    def test_sqrt(self):
        x = AD(4)
        f = el.sqrt(x)
        assert (f.val, f.der) == (2, 0.5 * 4 ** (-0.5))

        x = 4
        f = el.sqrt(x)
        assert (f.val, f.der) == (2, 0)

        x = 16.0
        f = el.sqrt(x)
        assert (f.val, f.der) == (4, 0)

        x = AD(-5)
        with pytest.raises(ValueError):
            f = el.sqrt(x)

        x = -5
        with pytest.raises(ValueError):
            f = el.sqrt(x)

        with pytest.raises(AttributeError):
            x = "World"
            f = el.sqrt(x)

        x = AD(4, N=3)
        f = el.sqrt(x)
        assert (f.val, f.der[-1]) == (2, (3 / 8) * 4 ** (-2.5))

    def test_sin(self):
        x = AD(0)
        f = el.sin(x)
        assert (f.val, f.der) == (0.0, 1.0)

        x = 13
        f = el.sin(x)
        assert (f.val, f.der) == (np.sin(13), 0)

        x = 13.0
        f = el.sin(x)
        assert (f.val, f.der) == (np.sin(13), 0)

        with pytest.raises(AttributeError):
            x = "!"
            f = el.sin(x)

    def test_cos(self):
        x = AD(90)
        f = el.cos(x)
        assert (f.val, f.der) == (np.cos(90), -np.sin(90))

        x = 13
        f = el.cos(x)
        assert (f.val, f.der) == (np.cos(13), 0)

        x = 13.0
        f = el.cos(x)
        assert (f.val, f.der) == (np.cos(13), 0)

        with pytest.raises(AttributeError):
            x = "ABC"
            f = el.cos(x)

    def test_tan(self):
        x = AD(90)
        f = el.tan(x)
        assert (f.val, f.der) == (np.tan(90), 1 / (np.cos(90)) ** 2)

        x = 13
        f = el.tan(x)
        assert (f.val, f.der) == (np.tan(13), 0)

        x = 13.0
        f = el.tan(x)
        assert (f.val, f.der) == (np.tan(13), 0)

        with pytest.raises(AttributeError):
            x = "xyz"
            f = el.tan(x)

    def test_arcsin(self):
        x = AD(0.5)
        f = el.arcsin(x)
        assert (f.val, f.der) == (np.arcsin(0.5), 1 / np.sqrt(1 - 0.5 ** 2))

        x = 0
        f = el.arcsin(x)
        assert (f.val, f.der) == (np.arcsin(0), 0)

        x = 1.0
        f = el.arcsin(x)
        assert (f.val, f.der) == (np.arcsin(1), 0)

        x = -5
        with pytest.raises(ValueError):
            f = el.arcsin(x)

        x = 10.0
        with pytest.raises(ValueError):
            f = el.arcsin(x)

        x = AD(-5.0)
        with pytest.raises(ValueError):
            f = el.arcsin(x)

        x = AD(10)
        with pytest.raises(ValueError):
            f = el.arcsin(x)

        with pytest.raises(AttributeError):
            x = "."
            f = el.arcsin(x)

    def test_arccos(self):
        x = AD(0.5)
        f = el.arccos(x)
        assert (f.val, f.der) == (np.arccos(0.5), -1 / np.sqrt(1 - 0.5 ** 2))

        x = 0.5
        f = el.arccos(x)
        assert (f.val, f.der) == (np.arccos(0.5), 0)

        x = -5
        with pytest.raises(ValueError):
            f = el.arccos(x)

        x = 10.0
        with pytest.raises(ValueError):
            f = el.arccos(x)

        x = AD(-5.0)
        with pytest.raises(ValueError):
            f = el.arccos(x)

        x = AD(10)
        with pytest.raises(ValueError):
            f = el.arccos(x)

        with pytest.raises(AttributeError):
            x = "--"
            f = el.arccos(x)

    def test_arctan(self):
        x = AD(0.5)
        f = el.arctan(x)
        assert (f.val, f.der) == (np.arctan(0.5), 1 / (1 + 0.5 ** 2))

        x = 0.5
        f = el.arctan(x)
        assert (f.val, f.der) == (np.arctan(0.5), 0)

        with pytest.raises(AttributeError):
            x = "AD(0.5)"
            f = el.arctan(x)

    def test_sinh(self):
        x = AD(0.5)
        f = el.sinh(x)
        assert (f.val, f.der) == (np.sinh(0.5), np.cosh(0.5))

        x = 0.5
        f = el.sinh(x)
        assert (f.val, f.der) == (np.sinh(0.5), 0)

        with pytest.raises(AttributeError):
            x = "0.5"
            f = el.sinh(x)

    def test_cosh(self):
        x = AD(0.5)
        f = el.cosh(x)
        assert (f.val, f.der) == (np.cosh(0.5), np.sinh(0.5))

        x = 0.5
        f = el.cosh(x)
        assert (f.val, f.der) == (np.cosh(0.5), 0)

        with pytest.raises(AttributeError):
            x = "0.5"
            f = el.cosh(x)

    def test_tanh(self):
        x = AD(0.5)
        f = el.tanh(x)
        assert (f.val, f.der) == (np.tanh(0.5), 1 - np.tanh(0.5) ** 2)

        x = 0.5
        f = el.tanh(x)
        assert (f.val, f.der) == (np.tanh(0.5), 0)

        with pytest.raises(AttributeError):
            x = "0.5"
            f = el.tanh(x)

    def test_add(self):
        x = AD(5)
        f = el.power(x, 2) + 5
        assert (f.val, f.der) == (30, 10)

        f = 5 + el.power(x, 2)
        assert (f.val, f.der) == (30, 10)

        f = el.power(x, 2) + 5 * x
        assert (f.val, f.der) == (50, 15)

        f = x * 5 + el.power(x, 2)
        assert (f.val, f.der) == (50, 15)

        with pytest.raises(AttributeError):
            f = el.power(x, 2) + "5"

        f = el.power(x, 2)
        g = 5
        h = f + g
        assert (h.val, h.der) == (30, 10)

        f = 5
        g = el.power(x, 2)
        h = f + g
        assert (h.val, h.der) == (30, 10)

    def test_sub(self):
        x = AD(5)
        f = el.power(x, 2) + -5 * x
        assert (f.val, f.der) == (0, 5)

        f = el.power(x, 2) - 50
        assert (f.val, f.der) == (-25, 10)

        f = - 50 + el.power(x, 2)
        assert (f.val, f.der) == (-25, 10)

        f = 50 - el.power(x, 2)
        assert (f.val, f.der) == (25, -10)

        f = -5 * x + el.power(x, 2)
        assert (f.val, f.der) == (0, 5)

        f = -x * 5 + el.power(x, 2)
        assert (f.val, f.der) == (0, 5)

        f = el.power(x, 2) - 5 * x
        assert (f.val, f.der) == (0, 5)

        f = x * 5 - el.power(x, 2)
        assert (f.val, f.der) == (0, -5)

        with pytest.raises(AttributeError):
            f = el.sin(x) - "5"

    def test_mul(self):
        x = AD(4)
        f = el.log(x, 2) * 3 ** x
        assert (np.round(f.val, 6), np.round(f.der[-1], 6)) == (
        162, np.round(81 / (4 * np.log(2)) + 162 * np.log(3), 6))

        f = 3 ** x * el.log(x, 2)
        assert (np.round(f.val, 6), np.round(f.der[-1], 6)) == (
        162, np.round(81 / (4 * np.log(2)) + 162 * np.log(3), 6))

        with pytest.raises(AttributeError):
            f = x * "5"

    def test_truediv(self):
        x = AD(4)
        f = el.log(x, 2) / 3 ** x
        assert (np.round(f.val, 6), np.round(f.der[-1], 6)) == (
        np.round(2 / 81, 6), np.round((81 / (4 * np.log(2)) - 162 * np.log(3)) / 3 ** 8, 6))

        f = el.sin(x) / 4
        assert (f.val, f.der) == ((np.sin(4)) / 4, (np.cos(4)) / 4)

        with pytest.raises(ZeroDivisionError):
            f = el.cos(x) / el.sin(0)

        with pytest.raises(ZeroDivisionError):
            f = el.cos(x) / 0

        f = 3 ** x / el.log(x, 2)
        assert (np.round(f.val, 6), np.round(f.der, 6)) == (np.round(81 / 2, 6),
                                                            np.round((162 * np.log(3) - 81 / (4 * np.log(2))) / (
                                                                        np.log(4) / np.log(2)) ** 2, 6))

        with pytest.raises(AttributeError):
            f = el.cos(x) / "el.sin(0)"

        x = AD(0)
        f = el.sin(x) / el.cos(x)
        assert (f.val, f.der) == (0, 1)

        with pytest.raises(ZeroDivisionError):
            f = el.cos(x) / el.sin(x)

    def test_pow(self):
        x = AD(2)
        f = x ** 4
        assert (f.val, f.der[-1]) == (16, 32)

        x = AD(2)
        f = 3 ** x
        assert (np.round(f.val, 8), np.round(f.der[-1], 6)) == (9, np.round(9 * np.log(3), 6))

        x = AD(2)
        f = el.power(x, 2)
        assert (f.val, f.der) == (4, 4)

        x = AD(2)
        f = x ** x
        assert (f.val, f.der[-1]) == (4, 4 * np.log(2) + 4)

        x = AD(2)
        f = (el.power(x, 2)) ** x
        assert (f.val, f.der) == (16, 16 * np.log(4) + 32)

        f = (el.power(x, 2)) ** 3
        assert (f.val, f.der) == (64, 192)

        with pytest.raises(AttributeError):
            f = (el.power(x, 2)) ** "3"

        f = (2 ** x) ** x
        assert (f.val, f.der) == (16, 16 * np.log(16))

        f = x ** (2 ** x)
        assert (f.val, f.der) == (16, 32 + 64 * (np.log(2) ** 2))

        x = AD(0)
        f = el.sin(x)
        g = x

        with pytest.raises(ValueError):
            h = f ** g

    def test_pos(self):
        x = AD(100)
        assert (x.val, x.der) == (100, 1)

        f = el.log(x, 10) + x
        assert (f.val, f.der) == (102, 1 / (100 * np.log(10)) + 1)

    def test_compare(self):
        n = 20
        x = AD(20)
        y = 100
        z = AD(100)

        assert (n < z.val) == n.__lt__(z.val)
        assert (x.val < y) == x.val.__lt__(y)

        assert (n <= z.val) == n.__le__(z.val)
        assert (x.val <= y) == x.val.__le__(y)
        assert (y <= z.val) == y.__le__(z.val)

        assert (y > x.val) == y.__gt__(x.val)
        assert (z.val > n) == z.val.__gt__(n)

        assert (y >= x.val) == y.__ge__(x.val)
        assert (z.val >= n) == z.val.__ge__(n)
        assert (n >= y) == n.__ge__(y)

        assert (y == z.val) == y.__eq__(z.val)
        assert (z.val == y) == z.val.__eq__(y)

        assert (y != x.val) == y.__ne__(x.val)
        assert (x.val != y) == x.val.__ne__(y)

        assert ("val = 20; der = [1.]") == x.__str__()





import sys, os
from SmartDiff.preprocess.pyexpr_formatter import PyExpression_Formatter

class TestPyExpFormatter:

  def test_init_default(self):
    pef = PyExpression_Formatter()
    assert pef.special_char_set == {"+", "-", "*", "/", ".", "(", ")", ",", "_", " "}

  def test_init_custom(self):
    char_set = {"+", "-", "*", "/", ".", "(", ")", ",", "{", "}", "[", "]", "_"}
    pef = PyExpression_Formatter(char_set)
    assert pef.special_char_set == char_set

  def test_is_valid_input_out0(self):
    pef = PyExpression_Formatter()
    s0 = "3 * (x + 8) - 3.4 / 1.750"
    s1 = "sqrt(48) * 3.9 + x"
    s2 = "log(x+4) / 20.4"
    s3 = "sin(x-2) ** 3"
    s4 = "arccos(cos(x))"
    s5 = "tanh(0.2*x_0)"
    s6 = "arcsin(sin(x_1/3)) + 0.8"
    s7 = "arctan(tan(x_2 + x_3))"
    s8 = "power(x_0123, 3 / 9)"
    s9 = "2.8 * exp(x + 4)"

    test_inputs = [s0, s1, s2, s3, s4, s5, s6, s7, s8, s9]
    for i in range(len(test_inputs)):
      s = test_inputs[i]
      assert pef.is_valid_input(s) == 0, "Test %d failed!" % i

  def test_is_valid_input_out0_with_e(self):
    pef = PyExpression_Formatter()
    s0 = "e**e"
    s1 = "e*2"
    s2 = "e / 3.89"
    s3 = "(e+8.4) - x * 10"
    s4 = "log(exp(e))"
    s5 = "arcsin(x) + e"
    s6 = "sin(e / 100.001) + sqrt(e)"
    s7 = "power(x, 3) + e / 2 "

    test_e_inputs = [s0, s1, s2, s3, s4, s5, s6, s7]
    for i in range(len(test_e_inputs)):
      s = test_e_inputs[i]
      assert pef.is_valid_input(s) == 0, "Test %d failed!" % i

  def test_is_valid_input_out0_with_pi(self):
    pef = PyExpression_Formatter()
    s0 = "pi**(2*pi)"
    s1 = "pi + 3 - 4*pi"
    s2 = "pi / 3.89"
    s3 = "(pi+8.4) - x * 10 - pi"
    s4 = "log(exp(pi)) + 8 "
    s5 = "arcsin(4.9*pi) + e"
    s6 = "cos( pi / 100.001) + sqrt( pi )"
    s7 = "power(pi, 3) + e / 2 "

    test_inputs = [s0, s1, s2, s3, s4, s5, s6, s7]
    for i in range(len(test_inputs)):
      s = test_inputs[i]
      assert pef.is_valid_input(s) == 0, "Test %d failed!" % i

  def test_is_valid_input_out1(self):
    # This set of inputs contain unmatched parenthesis
    pef = PyExpression_Formatter()
    s0 = "log(9) + ((x + 3)*8 + 9"
    s1 = "sinh(y) + )65 * (9 + 8)"
    s2 = "1 * (x + (e + 5 * (pi + 99)) / (x_2 + 3)"
    s3 = "exp((x + 9)"

    test_inputs = [s0, s1, s2, s3]
    for i in range(len(test_inputs)):
      s = test_inputs[i]
      assert pef.is_valid_input(s) == 1, "Test %d failed!" % i

  def test_is_valid_input_out2(self):
    pef = PyExpression_Formatter()
    s0 = "log(9) + [(x + 3)*8 + 9]"
    s1 = "sinh(y) + {65 * (9 + 8)}"
    s2 = "1 * (x + (e + 5 * (pi + 99))) / (x_2 + 3);"
    s3 = "3^4 + x"

    test_inputs = [s0, s1, s2, s3]
    for i in range(len(test_inputs)):
      s = test_inputs[i]
      assert pef.is_valid_input(s) == 2, "Test %d failed!" % i

  def test_format_to_pyexpr_no_e_no_pi(self):
    pef = PyExpression_Formatter()
    s0 = "3 * (x + 8) - 3.4 / 1.750 "
    s1 = "sqrt(48) * 3.9 + x"
    s2 = "log(x+4) / 20.4"
    s3 = "sin(x-2) ** 3"
    s4 = "arccos(cos(x))"
    s5 = "tanh(0.2*x_0)"
    s6 = "arcsin(sin(x_1/3)) + 0.8 "
    s7 = "arctan(tan(x_2 + x_3))"
    s8 = "power(x_0123, 3 / 9) "
    s9 = "2.8 * exp(x + 4)"
    test_inputs = [s0, s1, s2, s3, s4, s5, s6, s7, s8, s9]
    expected_outs = [s.replace(" ", "") for s in test_inputs]

    for i in range(len(test_inputs)):
      s = test_inputs[i]
      assert pef.format_to_pyexpr(s) == expected_outs[i], "Test %d failed!" % i

  def test_format_to_pyexpr_with_e(self):
    pef = PyExpression_Formatter()
    s0 = "2.5 * e**e"
    s1 = "e / 3.89"
    s2 = "(e+8.4) - x * 10"
    s3 = "log(exp(e))"
    s4 = "arcsin(x) + e"
    s5 = "sin(e / 100.001) + sqrt(e)"
    s6 = "power(x, 3) + e / 2 "
    test_inputs = [s0, s1, s2, s3, s4, s5, s6]

    a0 = "2.5*math.e**math.e"
    a1 = "math.e/3.89"
    a2 = "(math.e+8.4)-x*10"
    a3 = "log(exp(math.e))"
    a4 = "arcsin(x)+math.e"
    a5 = "sin(math.e/100.001)+sqrt(math.e)"
    a6 = "power(x,3)+math.e/2"
    expected_outs = [a0, a1, a2, a3, a4, a5, a6]

    for i in range(len(test_inputs)):
      s = test_inputs[i]
      assert pef.format_to_pyexpr(s) == expected_outs[i], "Test %d failed!" % i

  def test_format_to_pyexpr_with_pi(self):
    pef = PyExpression_Formatter()
    s0 = "pi**(2*pi)"
    s1 = "pi + 3 - 4*pi"
    s2 = "pi / 3.89"
    s3 = "(pi+8.4) - x * 10 - pi"
    s4 = "log(exp(pi)) + 8 "
    s5 = "arcsin(4.9*pi) + e"
    s6 = "cos( pi / 100.001) + sqrt( pi )"
    s7 = "power(pi, 3) + e / 2 "
    test_inputs = [s0, s1, s2, s3, s4, s5, s6, s7]

    a0 = "math.pi**(2*math.pi)"
    a1 = "math.pi+3-4*math.pi"
    a2 = "math.pi/3.89"
    a3 = "(math.pi+8.4)-x*10-math.pi"
    a4 = "log(exp(math.pi))+8"
    a5 = "arcsin(4.9*math.pi)+math.e"
    a6 = "cos(math.pi/100.001)+sqrt(math.pi)"
    a7 = "power(math.pi,3)+math.e/2"
    test_outputs = [a0, a1, a2, a3, a4, a5, a6, a7]

    for i in range(len(test_inputs)):
      s = test_inputs[i]
      assert pef.format_to_pyexpr(s) == test_outputs[i], "Test %d failed!\n" % i



import numpy as np
from sympy import symbols, ordered, Matrix, hessian
from sympy.core.sympify import sympify

from SmartDiff.globals import FUNC_MAP, MATH_FUNC_MAP
from SmartDiff.solvers.element_op import AutoDiff as AD
from SmartDiff.solvers.element_op import *


def get_ord2_der(func_str, all_vals, dvar_idx, func_map):
    """
    Returns the 2nd order derivative of dvar_idx^th variable
    :param func_str: A string of input math function
    :param all_vals: A list of real scalar values (in the same order as the variable names)
    :param dvar_idx: Index of the variable to take derivative
    :param func_map: A mapping of math expression to python's math function
    :return:
    """
    var_map = {"x%d" % dvar_idx: AutoDiff(value=all_vals[dvar_idx], N=2)}
    var_map.update({"x%d" % idx: val for idx, val in enumerate(all_vals) if idx != dvar_idx})
    var_map.update(func_map)
    AD_out = eval(func_str, var_map)
    return AD_out.der[-1]


def get_hessian(func_str, all_vals, eval_func_map=FUNC_MAP, math_func_map=MATH_FUNC_MAP):
    """
    Returns the hessian matrix of the input function over the input variables
    :param func_str: A string of input math function
    :param all_vals: A list of real scalar values
    :param eval_func_map: A mapping of math expression to python's math function
    :param math_func_map: A mapping of math expression to python's math function
    :return:
    """
    # Assume func is a valid expression
    D = len(all_vals)
    assert D > 0, "There should be at least 1 variable!"

    H = np.zeros((D, D))
    if D == 1:
        H[0][0] = get_ord2_der(func_str, all_vals, 0, eval_func_map)
    else:
        var_map = {"x%d" % i: val for i, val in enumerate(all_vals)}
        var_map.update(math_func_map)
        f = sympify(func_str)
        vs = f.free_symbols
        hess = hessian(f, list(ordered(vs)))
        # print(hess)
        for i in range(D):
            for j in range(D):
                didj_func = hess[i * D + j]
                H[i][j] = eval(str(didj_func), var_map)
    return H


def get_val_jacobian(func_list, vals, eval_func_map=FUNC_MAP):
    '''
    Calulates the function values and Jacobian for multivariate input
    :param func_list: a list of m function strs
    :param vals: list of scalar, len = n
    :param eval_func_map: A mapping of math expression to python's math function
    :return: value, Jacobian
    '''
    n = len(vals)
    m = len(func_list)
    jacobian = np.zeros((m, n))
    values = np.zeros(m)
    xs = symbols('x:%d' % n)
    for i, f in enumerate(func_list):
        var_map = {str(xs[i]): val for i, val in enumerate(vals)}
        var_map.update(eval_func_map)
        get_val = False
        for j, x in enumerate(vals):
            AD_map = var_map.copy()
            AD_map.update({str(xs[j]): AutoDiff(value=vals[j])})
            AD_out = eval(f, AD_map)
            if not get_val:
                try:
                    values[i] = AD_out.val
                    get_val = True
                except AttributeError:
                    pass
            try:
                jacobian[i, j] = AD_out.der[0]
            except AttributeError:
                pass

    return values, jacobian




import numpy as np
from sympy import bell, binomial, symbols
from math import factorial

# Here contains the derivative calculation of elementary operations
# If the input 'x' is a scalar number, it will return the value of such operation evaluated at 'x',
# which is directly applied the operation.
# If the input is a dual number(AutoDiff object), it will return another dual number(AutoDiff object),
# where the val represents the value evaluated at 'x', der represents the derivative to 'x' which evaluated at 'x'.
# derivatives for reference: http://math2.org/math/derivatives/tableof.html


class AutoDiff():
    """
    The AutoDiff class defines a dual number which return the value and N^th order derivative of a given function
    User can give a variable and order of the derivative they want to differentiate in this class.
    """

    def __init__(self, value, der=None, N=1):
        """
        Initiate the variable dual object with given value, derivative, order of derivative
        INPUTS
        =======
        value: a scalar, represent the number user want to evaluate at.
        der: a scalar or a list, represent the derivative of input variable.
            If no der is given, it will get the value 1 by default.
        N: a integer, represent the order of derivative.
            If no N is given, it will get the value 1 by default.
        """
        self.val = value
        self.N = N
        if der is None:
            self.der = np.zeros(N)
            self.der[0] = 1.0
        else:
            self.der = np.array(der)

    def __add__(self, other):
        # (f+g)' = f' + g'
        # (f+g)^(n) = f^(n) + g^(n)
        """
        Return a new AutoDiff object, which is a sum of given AutoDiff object and a new AutoDiff object.
        INPUTS
        =======
        self: an AutoDiff object, represent the given AutoDiff object before '+' operation.
        other: an AutoDiff object or a real number (float/int), represent the new part after '+' operation.
        RETURNS
        =======
        a new AutoDiff object after the '+' calculation.
        If N is given, the der of AutoDiff object will be a list of derivative from order 0 to order N.
        EXAMPLES
        =======
        >>> a = AutoDiff(5)
        >>> b = AutoDiff(6)
        >>> f = a + b
        >>> print(f.val, f.der)
        11 [2.]
        >>> a = AutoDiff(5)
        >>> f = a + 2
        >>> print(f.val, f.der)
        7 [1.]
        >>> a = AutoDiff(5, N = 3)
        >>> b = AutoDiff(6, N = 3)
        >>> f = a + b
        >>> print(f.val, f.der)
        11 [2. 0. 0.]
        """
        try:
            val_new = self.val + other.val
            N_new = self.N  # The larger order of the two components
            if self.N == other.N:
                der_new = self.der + other.der
            elif self.N > other.N:  # other is constant
                der_new = self.der
            else:  # self is constant
                N_new = other.N
                der_new = other.der
        except AttributeError:
            if isinstance(other, float) or isinstance(other, int):
                val_new = self.val + other
                der_new = self.der
                N_new = self.N
            else:
                raise AttributeError('Type error!')
        return AutoDiff(val_new, der_new, N_new)

    def __radd__(self, other):
        """
        Return a new AutoDiff object, which is a sum of new AutoDiff object and the given AutoDiff object.
        Implement a reverse AutoDiffdition of __add__()
        EXAMPLES
        =======
        >>> a = AutoDiff(5)
        >>> f = 2 + a
        >>> print(f.val, f.der)
        7 [1.]
        """
        return self.__add__(other)

    def __sub__(self, other):
        # (f-g)' = f' - g'
        # (f-g)^(n) = f^(n) - g^(n)
        """
        Return a new AutoDiff object, which is a subtraction of given AutoDiff object and a new AutoDiff object.
        INPUTS
        =======
        self: an AutoDiff object, represent the given AutoDiff object before '-' operation.
        other: an AutoDiff object or a real number (float/int), represent the new part after '-' operation.
        RETURNS
        =======
        a new AutoDiff object after the '-' calculation.
        If N is given, the der of AutoDiff object will be a list of derivative from order 0 to order N.
        EXAMPLES
        =======
        >>> a = AutoDiff(5)
        >>> b = AutoDiff(6)
        >>> f = a - b
        >>> print(f.val, f.der)
        -1 [0.]
        >>> a = AutoDiff(5)
        >>> f = a - 2
        >>> print(f.val, f.der)
        3 [1.]
        >>> a = AutoDiff(5, N = 3)
        >>> b = AutoDiff(6, N = 3)
        >>> f = a - b
        >>> print(f.val, f.der)
        -1 [0. 0. 0.]
        """
        try:
            val_new = self.val - other.val
            N_new = self.N  # The larger order of the two components
            if self.N == other.N:
                der_new = self.der - other.der
            elif self.N > other.N:  # other is constant
                der_new = self.der
            else:  # self is constant
                N_new = other.N
                der_new = other.der
        except AttributeError:
            if isinstance(other, float) or isinstance(other, int):
                val_new = self.val - other
                der_new = self.der
                N_new = self.N
            else:
                raise AttributeError('Type error!')
        return AutoDiff(val_new, der_new, N_new)

    def __rsub__(self, other):
        # In this case, other must be a constant
        """
        Return a new AutoDiff object, which is a subtraction of new AutoDiff object and the given AutoDiff object.
        Implement a reverse subtraction of __sub__().
        EXAMPLES
        =======
        >>> a = AutoDiff(5)
        >>> f = 2 - a
        >>> print(f.val, f.der)
        -3 [-1.]
        """
        val_new = other - self.val
        der_new = -self.der
        return AutoDiff(val_new, der_new, self.N)

    def __mul__(self, other):
        # (f*g)' = f'*g + g' * f
        # (f*g)^(n) = sum_{k=0}^{n} [binom(n, k) * f^(n-k)g^(k)]
        """
        Return a new AutoDiff object, which is a product of given AutoDiff object and a new AutoDiff object.
        INPUTS
        =======
        self: an AutoDiff object, represent the given AutoDiff object before '*' operation.
        other: an AutoDiff object or a real number (float/int), represent the new part after '*' operation.
        RETURNS
        =======
        a new AutoDiff object after the '*' calculation.
        If N is given, the der of AutoDiff object will be a list of derivative from order 0 to order N.
        EXAMPLES
        =======
        >>> a = AutoDiff(5)
        >>> b = AutoDiff(6)
        >>> f = a * b
        >>> print(f.val, f.der)
        30 [11.]
        >>> a = AutoDiff(5)
        >>> f = a * 2
        >>> print(f.val, f.der)
        10 [2.]
        >>> a = AutoDiff(5, N = 3)
        >>> b = AutoDiff(6, N = 3)
        >>> f = a * b
        >>> print(f.val, f.der)
        30 [0. 2.0 0.]
        """
        try:
            val_new = self.val * other.val
            N_new = self.N
            if self.N == other.N:
                der_new = []
                for n in range(1, self.N+1):
                    # binomial(n, 0) = binomial(n, n) = 1 always holds
                    nth_der = self.der[-1] * other.val + self.val * other.der[-1]  # 1st and last term
                    for k in range(1, n):
                        nth_der += binomial(n, k) * self.der[n-k-1] * other.der[k-1]
                    der_new.append(nth_der)
            elif self.N > other.N:  # other is a constant
                der_new = self.der * other.val
            else:   # self is a constant
                der_new = other.der * self.val
                N_new = other.N

        except AttributeError:
            if isinstance(other, float) or isinstance(other, int):
                val_new = self.val * other
                der_new = self.der * other  # other is a constant in this case
                N_new = self.N
            else:
                raise AttributeError('Type error!')
        return AutoDiff(val_new, der_new, N_new)

    def __rmul__(self, other):
        """
        Return a new AutoDiff object, which is a product of new AutoDiff object and the given AutoDiff object.
        Implement a reverse product of __mul__().
        EXAMPLES
        =======
        >>> a = AutoDiff(5)
        >>> f = 2 * a
        >>> print(f.val, f.der)
        10 [2. 0. 0.]
        """
        return self.__mul__(other)

    def __truediv__(self, other):
        # (f/g)' = (f'*g - g'*f)/g^2
        """
        Return a new AutoDiff object, which is a division of given AutoDiff object and a new AutoDiff object.
        INPUTS
        =======
        self: an AutoDiff object, represent the given AutoDiff object before '/' operation.
        other: an AutoDiff object or a real number (float/int), represent the new part after '/' operation.
        RETURNS
        =======
        a new AutoDiff object after the '/' calculation.
        If N is given, the der of AutoDiff object will be a list of derivative from order 0 to order N.
        EXAMPLES
        =======
        >>> a = AutoDiff(6)
        >>> b = AutoDiff(2)
        >>> f = a / b
        >>> print(f.val, f.der)
        3.0 [-1.]
        >>> a = AutoDiff(6)
        >>> f = a / 2
        >>> print(f.val, f.der)
        3.0 [0.5]
        >>> a = AutoDiff(6, N = 3)
        >>> b = AutoDiff(2, N = 3)
        >>> f = a / b
        >>> print(f.val, f.der)
        3.0 [-2.25 -2.75 -1.5]
        """
        try:
            if other.val != 0:
                # elementwise division to return Jacobian matrix
                denom = inv(other)
                return self.__mul__(denom)
            else:
                raise ZeroDivisionError('Division by zero')
        except AttributeError:
            if isinstance(other, float) or isinstance(other, int):
                if other != 0:
                    # just divide if other is constant
                    val_new = self.val / other
                    der_new = self.der / other
                    return AutoDiff(val_new, der_new, self.N)
                else:
                    raise ZeroDivisionError('Division by zero')
            else:
                raise AttributeError('Type error!')

    def __rtruediv__(self, other):
        """
        Return a new AutoDiff object, which is a division of new AutoDiff object and the given AutoDiff object.
        Implement a reverse product of __truediv__().
        EXAMPLES
        =======
        >>> a = AutoDiff(5)
        >>> f = 2 / a
        >>> print(f.val, f.der)
        0.4 [-0.08]
        >>> a = AutoDiff(5, N =3)
        >>> f = 2 / a
        >>> print(f.val, f.der)
        0.4 [-0.08 0.032 -0.0192]
        """
        if self.val != 0:
            inv_self = inv(self)
            return inv_self * other
        else:
            raise ZeroDivisionError('Division by zero')

    def __pow__(self, other):
        # (f^g)' = f^g * (f'/f * g + g' * ln(f))
        """
        Return a new AutoDiff object, which is a power calculation of given AutoDiff object and a new AutoDiff object.
        INPUTS
        =======
        self: an AutoDiff object, represent the given AutoDiff object for the base of power.
        other: an AutoDiff object or a real number (float/int), represent the exponent part of power.
        RETURNS
        =======
        a new AutoDiff object after the self**(other) power calculation.
        If N is given, the der of AutoDiff object will be a list of derivative from order 0 to order N.
        EXAMPLES
        =======
        >>> a = AutoDiff(6)
        >>> b = AutoDiff(2)
        >>> f = a ** b
        >>> print(f.val, f.der)
        36.0 [76.50334089]
        >>> a = AutoDiff(6)
        >>> f = a ** 2
        >>> print(f.val, f.der)
        36.0 [12.]
        >>> a = AutoDiff(6, N = 3)
        >>> b = AutoDiff(2, N = 3)
        >>> f = a / b
        >>> print(f.val, f.der)
        36.0 [0.6666666666666666 12.6790123456790 -1.62940100594422]
        """
        if self.val <= 0:
            raise ValueError('Error: Value of base function must be positive!')
        try:
            # elementwise power to return Jacobian matrix
            val_new = self.val ** other.val
            N_new = self.N  # The larger order of the two components
            if self.N == other.N:
                if N_new == 1:
                    der_new = val_new * (other.val * self.der / self.val + other.der * np.log(self.val))
                else:
                    # fx^gx = e^(gx * ln(fx))
                    pw = ln(self) * other
                    f_power_g = exp(pw)
                    return f_power_g
            elif self.N > other.N:
                # other is constant: x^a
                return power(self, other.val)
            else:
                # self is constant: b^gx = e^(gx * ln(b))
                pw = other * np.log(self.val)
                return exp(pw)
        except AttributeError:
            if isinstance(other, float) or isinstance(other, int):
                return power(self, other)
            else:
                raise AttributeError('Type error!')
        return AutoDiff(val_new, der_new, N_new)

    def __rpow__(self, other):
        # other is constant: c^fx = e^(fx * ln(c))
        """
        Return a new AutoDiff object, which is a power calculation of new AutoDiff object and the given AutoDiff object.
        Implement a reverse product of __pow__().
        EXAMPLES
        =======
        >>> a = AutoDiff(2)
        >>> f = 2 ** a
        >>> print(f.val, f.der)
        4.0 [2.77258872]
        >>> a = AutoDiff(2, N =3)
        >>> f = 2 ** a
        >>> print(f.val, f.der)
        4.0 [2.77258872 1.92181206 1.33209861]
        """
        pw = self * np.log(other)
        return exp(pw)

    # unary operations
    def __neg__(self):
        """
        Return a new AutoDiff object, which is a negative unary operation for the given AutoDiff object.
        INPUTS
        =======
        self: an AutoDiff object, represent the given AutoDiff object for the negative unary operation.
        RETURNS
        =======
        a new AutoDiff object after the negation operation.
        EXAMPLES
        =======
        >>> a = AutoDiff(2)
        >>> f = -a
        >>> print(f.val, f.der)
        -2 [-1.]
        """
        val_new = -self.val
        der_new = -self.der
        return AutoDiff(val_new, der_new, self.N)

    def __pos__(self):
        """
        Return a new AutoDiff object, which is a positive unary operation for the given AutoDiff object.
        INPUTS
        =======
        self: an AutoDiff object, represent the given AutoDiff object for the positive unary operation.
        RETURNS
        =======
        a new AutoDiff object after the negation operation.
        EXAMPLES
        =======
        >>> a = AutoDiff(2)
        >>> f = a
        >>> print(f.val, f.der)
        2 [1.]
        """
        val_new = self.val
        der_new = self.der
        return AutoDiff(val_new, der_new, self.N)

    # comparison operator
    def __lt__(self, other):
        """
        Return the result of less than comparison operator.
        INPUTS
        =======
        self: an AutoDiff object or number (float/int), represent the given AutoDiff object for comparison before.
        other: an AutoDiff object or number (float/int), represent the given AutoDiff object for comparison after.
        RETURNS
        =======
        The boolean result of less than comparison.
        EXAMPLES
        =======
        >>> x = AutoDiff(1)
        >>> y = AutoDiff(2)
        >>> x < y
        True
        """
        try:
            return self.val < other.val
        except AttributeError:
            return self.val < other

    def __gt__(self, other):
        """
        Return the result of greater than comparison operator.
        INPUTS
        =======
        self: an AutoDiff object or number (float/int), represent the given AutoDiff object for comparison before.
        other: an AutoDiff object or number (float/int), represent the given AutoDiff object for comparison after.
        RETURNS
        =======
        The boolean result of greater than comparison.
        EXAMPLES
        =======
        >>> x = AutoDiff(1)
        >>> y = AutoDiff(2)
        >>> x >= y
        False
        """
        try:
            return self.val > other.val
        except AttributeError:
            return self.val > other


    def __le__(self, other):
        """
        Return the result of less than or equal to comparison operator.
        INPUTS
        =======
        self: an AutoDiff object or number (float/int), represent the given AutoDiff object for comparison before.
        other: an AutoDiff object or number (float/int), represent the given AutoDiff object for comparison after.
        RETURNS
        =======
        The boolean result of less than or equal to comparison.
        EXAMPLES
        =======
        >>> x = AutoDiff(1)
        >>> y = AutoDiff(2)
        >>> x <= y
        True
        """
        try:
            return self.val <= other.val
        except AttributeError:
            return self.val <= other

    def __ge__(self, other):
        """
        Return the result of greater than or equal to comparison operator.
        INPUTS
        =======
        self: an AutoDiff object or number (float/int), represent the given AutoDiff object for comparison before.
        other: an AutoDiff object or number (float/int), represent the given AutoDiff object for comparison after.
        RETURNS
        =======
        The boolean result of greater than or equal to comparison.
        EXAMPLES
        =======
        >>> x = AutoDiff(2)
        >>> y = AutoDiff(1)
        >>> x >= y
        True
        """
        try:
            return self.val >= other.val
        except AttributeError:
            return self.val >= other

    def __eq__(self, other):
        """
        Return the result of equal to comparison operator.
        INPUTS
        =======
        self: an AutoDiff object or number (float/int), represent the given AutoDiff object for comparison before.
        other: an AutoDiff object or number (float/int), represent the given AutoDiff object for comparison after.
        RETURNS
        =======
        The boolean result of equal to comparison.
        EXAMPLES
        =======
        >>> x = AutoDiff(1)
        >>> y = AutoDiff(2)
        >>> x == y
        False
        """
        try:
            return self.val == other.val
        except AttributeError:
            return self.val == other


    def __ne__(self, other):
        """
        Return the result of not equal to comparison operator.
        INPUTS
        =======
        self: an AutoDiff object or number (float/int), represent the given AutoDiff object for comparison before.
        other: an AutoDiff object or number (float/int), represent the given AutoDiff object for comparison after.
        RETURNS
        =======
        The boolean result of not equal to comparison.
        EXAMPLES
        =======
        >>> x = AutoDiff(1)
        >>> y = AutoDiff(2)
        >>> x != y
        True
        """
        try:
            return not self.val == other.val
        except AttributeError:
            return not self.val == other

    def __str__(self):
        val = "val = " + str(self.val)
        der = "; der = " + str(self.der)
        return val + der



def get_n_der_vecs(dk_f, gx, N):
    """
    This function applies Faa di Bruno's formula to compute the derivatives of order from 1 to N
    given the calling elementary operator has dk_f as its kth order derivative function.
    INPUTS
    =======
    dk_f(val, k): A lambda function of the kth order derivative of f at the point val
    gx: Potentially an AutoDiff object
    N: highest derivative order of gx
    RETURNS
    =======
    a list of high-order derivatives up until gx.N
    """
    # Create symbols and symbol-value mapping for eval() in the loop
    dxs = symbols('q:%d' % N)
    dx_mapping = {str(dxs[i]): gx.der[i] for i in range(N)}
    # Use Faa di Bruno's formula
    der_new = []
    for n in range(1, N + 1):
        nth_der = 0
        for k in range(1, n + 1):
            # The first n-k+1 derivatives
            t = n - k + 1
            vars = dxs[:t]
            # bell polynomial as python function str
            bell_nk_str = str(bell(n, k, vars))
            # evaluate the bell polynomial using the symbol-value mapping
            val_bell_nk = eval(bell_nk_str, dx_mapping)
            nth_der += dk_f(gx.val, k) * val_bell_nk
        der_new.append(nth_der)
    return der_new

def power_k_order(gx, n, k):
    """
    Returns the kth order derivative of gx**n
    INPUTS
    =======
    gx: Potentially an AutoDiff object
    n: float or int, required, the base of power
    k: highest derivative order
    RETURNS
    =======
    new AutoDiff object after gx**n calculation
    """
    if type(n) != int and type(n) != float:
        raise AttributeError('Type error!')

    if n == 1/2 and gx < 0:
        raise ValueError('Error: Independent variable must be nonnegative!')

    if n == int(n) and n >= 0:
        if k <= n:
            falling = 1
            for i in range(k):
                falling *= n - i
            return falling * gx ** (n-k)
        if k > n:
            return 0
    else:
        falling = 1
        for i in range(k):
            falling *= n - i
        return falling * gx ** (n - k)

def power(x, n):
    # ((x)^n)' = n * (x)^{n-1} * x'
    """
    Returns the value and derivative of a power operation: x**n
    INPUTS
    =======
    x: an AutoDiff object or a scalar, required, the input variable
    n: float or int, required, the base
    RETURNS
    ========
    an AutoDiff object containing the value and derivative of the expression
    EXAMPLES
    =========
    >>> f = power(1.0, 2.0)
    >>> print(f.val, f.der)
    1.0 [0.]
    >>> f = power(AutoDiff(1.0, N=3), 2.0)
    >>> print(f.val, f.der)
    1.0, [2. 2. 0.]
    """
    N = 1
    dk_f = lambda gx, k: power_k_order(gx, n, k) # nth order derivative for power(x,n)
    try:
        val_new = np.power(x.val, n)
        N = x.N
        if N == 1:
            der_new = n * x.val ** (n - 1) * x.der
        else:
            # N > 1
            der_new = get_n_der_vecs(dk_f, x, N)
    except AttributeError:
        if isinstance(x, float) or isinstance(x, int):
            val_new = np.power(x, n)
            # If x is a constant, the derivative of x is 0.
            der_new = np.zeros(1)
        else:
            raise AttributeError('Type error!')
    return AutoDiff(val_new, der_new, N)


def log(x, a):
    # (log_a(x))' = 1/(x * log_e(a) * x')
    # we should also check the value >0 for log calculation
    # equivalent to log_gx(fx) = ln(fx) / ln(gx)
    """
    Returns the value and derivative of a logarithm operation: log_n(x)
    INPUTS
    =======
    x: an AutoDiff object or a scalar, required, the input variable
    a: float or int, required, the base
    RETURNS
    ========
    an AutoDiff object containing the value and derivative of the expression
    EXAMPLES
    =========
    >>> f = log(np.e, np.e)
    >>> print(f.val, f.der)
    1.0 [0.]
    >>> f = log(AutoDiff(np.e**2, N = 3), np.e)
    >>> print(f.val, f.der)
    2.0 [0.13533528 -0.01831564 0.0049575]
    """
    if isinstance(x, AutoDiff):
        if x.val <= 0:
            raise ValueError('Error: Independent variable must be positive!')
    try:
        return ln(x) / ln(a)
    except AttributeError:
        if isinstance(x, float) or isinstance(x, int):
            if x <= 0:
                raise ValueError('Error: Independent variable must be positive!')
            return inv(ln(a)) * np.log(x)  # in case a is an AutoDiff object
        else:
            raise AttributeError('Type error!')


def ln(x):
    # (log_n(x))' = 1/(x * log_e(n) * x')
    # we should also check the value >0 for log calculation
    """
    Returns the value and derivative of a logarithm operation: log_n(x)
    INPUTS
    =======
    x: an AutoDiff object or a scalar, required, the input variable
    a: float or int, required, the base
    RETURNS
    ========
    an AutoDiff object containing the value and derivative of the expression
    EXAMPLES
    =========
    >>> f = ln(np.e**2)
    >>> print(f.val, f.der)
    2.0 [0.]
    >>> f = ln(AutoDiff(np.e**2, N=3))
    >>> print(f.val, f.der)
    2.0 [0.13533528 -0.01831564 0.0049575]
    """
    if isinstance(x, AutoDiff):
        if x.val <= 0:
            raise ValueError('Error: Independent variable must be positive!')
    N = 1
    dk_f = lambda gx, k: ((-1.0)**(k-1) * factorial(k-1)) / (gx ** k)
    try:
        val_new = np.log(x.val)
        N = x.N
        if N == 1:
            der_new = x.der * (1 / x.val)
        else:
            der_new = get_n_der_vecs(dk_f, x, N)
    except AttributeError:
        if isinstance(x, float) or isinstance(x, int):
            if x <= 0:
                raise ValueError('Error: Independent variable must be positive!')
            val_new = np.log(x)
            # If x is a constant, the derivative of x is 0.
            der_new = np.zeros(1)
        else:
            raise AttributeError('Type error!')
    return AutoDiff(val_new, der_new, N)


def expn(x,n):
    #  (n^{x})' = n^{x} * ln(n) * x'
    """
    Returns the value and derivative of a exponential operation: n^x
    INPUTS
    =======
    x: an AutoDiff object or a scalar, required, the input variable
    n: float or int, required, the base
    RETURNS
    ========
    an AutoDiff object containing the value and derivative of the expression
    EXAMPLES
    =========
    >>> f = expn(1.0, np.e)
    >>> print(f.val, f.der)
    2.718281828459045 [0.]
    >>> f = expn(AutoDiff(1.0, N=3), np.e)
    >>> print(f.val, f.der)
    2.718281828459045 [2.71828183 2.71828183 2.71828183])
    """
    if isinstance(n, float) or isinstance(n, int):
        if n < 0:
            raise ValueError('Error: Base must be positive!')
    N = 1
    dk_f = lambda gx, k: n**gx *(np.log(n)**2)  # nth order derivative for n^x
    try:
        val_new = n**x.val
        N = x.N
        if N == 1:
            der_new = n**x.val * np.log(n) * x.der
        else:
            # N > 1
            der_new = get_n_der_vecs(dk_f, x, N)
    except AttributeError:
        if isinstance(x, float) or isinstance(x, int):
            val_new = n**x
            # If x is a constant, the derivative of x is 0.
            der_new = np.zeros(1)
        else:
            raise AttributeError('Type error!')
    return AutoDiff(val_new, der_new, N)


def exp(x):
    # (e^{x})' = e^{x} * x'
    """
    Returns the value and derivative of a exponential operation: e^x
    INPUTS
    =======
    x: an AutoDiff object or a scalar, required, the input variable
    RETURNS
    ========
    an AutoDiff object containing the value and derivative of the expression
    EXAMPLES
    =========
    >>> f = exp(1.0)
    >>> print(f.val, f.der)
    2.718281828459045 [0.]
    >>> f = exp(AutoDiff(1.0, N=3))
    >>> print(f.val, f.der)
    2.718281828459045 [2.71828183 2.71828183 2.71828183]
    """
    N = 1
    dk_f = lambda gx, k: np.exp(gx)  # nth order derivative for e^x
    try:
        val_new = np.exp(x.val)
        N = x.N
        if N == 1:
            der_new = np.exp(x.val) * x.der
        else:
            # N > 1
            der_new = get_n_der_vecs(dk_f, x, N)
    except AttributeError:
        if isinstance(x, float) or isinstance(x, int):
            val_new = np.exp(x)
            # If x is a constant, the derivative of x is 0.
            der_new = np.zeros(1)
        else:
            raise AttributeError('Type error!')
    return AutoDiff(val_new, der_new, N)


def inv(x):
    """
    Inverse of a term (x cannot be 0)
    :param x:
    :return:
    """
    N = 1
    dk_f = lambda gx, k: ((-1.0) ** k * factorial(k)) / (gx ** (k+1)) # YW
    try:
        val_new = 1.0 / x.val
        N = x.N
        if N == 1:
            der_new = -x.der / (x.val**2)
        else:
            der_new = get_n_der_vecs(dk_f, x, N)
    except AttributeError:
        if isinstance(x, float) or isinstance(x, int):
            if x == 0:
                raise ZeroDivisionError('Division by zero')
            val_new = 1.0 / x
            # If x is a constant, the derivative of x is 0.
            der_new = np.zeros(1)
        else:
            raise AttributeError('Type error!')
    return AutoDiff(val_new, der_new, N) # YW


def sqrt(x):
    # (sqrt(x))' = ((x)^{1/2})' = 1/2 * (x)^{-1/2} * x'
    # we should also check the value is >0 for sqrt calculation
    """
    Returns the value and derivative of a square root operation: x^{1/2}
    INPUTS
    =======
    x: an AutoDiff object or a scalar, required, the input variable
    RETURNS
    ========
    an AutoDiff object containing the value and derivative of the expression
    EXAMPLES
    =========
    >>> f = sqrt(1.0)
    >>> print(f.val, f.der)
    1.0 [0.]
    >>> f = sqrt(AutoDiff(1.0, N = 3))
    >>> print(f.val, f.der)
    1.0 [0.5 -0.25 0.375]
    """
    if isinstance(x, AutoDiff):
        if x.val < 0:
            raise ValueError('Error: Independent variable must be nonnegative!')
    N = 1
    dk_f = lambda gx, k: power_k_order(gx, 1/2, k)  # nth order derivative for sqrt(x)
    try:
        val_new = np.sqrt(x.val)
        N = x.N
        if N == 1:
            # der_new = np.array([1 / 2 * x.val ** (-1/2) * x.der])
            der_new = 0.5 * x.der / np.sqrt(x.val)
        else:
            # N > 1
            der_new = get_n_der_vecs(dk_f, x, N)
    except AttributeError:
        if isinstance(x, float) or isinstance(x, int):
            if x < 0:
                raise ValueError('Error: Independent variable must be nonnegative!')
            val_new = np.sqrt(x)
            # If x is a constant, the derivative of x is 0.
            der_new = np.zeros(1)
        else:
            raise AttributeError('Type error!')
    return AutoDiff(val_new, der_new, N)


def sin(x):
    # (sin(x))' = cos(x) * x'
    """
    Returns the value and derivative of a sine operation: sin(x)
    INPUTS
    =======
    x: an AutoDiff object or a scalar, required, the input variable
    RETURNS
    ========
    an AutoDiff object containing the value and derivative of the expression
    EXAMPLES
    =========
    >>> f = sin(0.0)
    >>> print(f.val, f.der)
    0.0 [0.]
    >>> f = sin(AutoDiff(0.0, N=3))
    >>> print(f.val, f.der)
    0.0 [1.0000000e+00 1.2246468e-16 -1.0000000e+00]
    """
    half_pi = np.pi / 2
    N = 1
    dk_f = lambda gx, k: np.sin(gx + half_pi * k)  # nth order derivative for sin(x)
    try:
        val_new = np.sin(x.val)
        N = x.N
        if N == 1:
            der_new = np.cos(x.val) * x.der
        else:
            # N > 1
            der_new = get_n_der_vecs(dk_f, x, N)
    except AttributeError:
        if isinstance(x, float) or isinstance(x, int):
            val_new = np.sin(x)
            # If x is a constant, the derivative of x is 0.
            der_new = np.zeros(1)
        else:
            raise AttributeError('Type error!')
    return AutoDiff(val_new, der_new, N)


def cos(x):
    # (cos(x))' = - sin(x) * x'
    """
    Returns the value and derivative of a cosine operation: cos(x)
    INPUTS
    =======
    x: an AutoDiff object or a scalar, required, the input variable
    RETURNS
    ========
    an AutoDiff object containing the value and derivative of the expression
    EXAMPLES
    =========
    >>> f = cos(0.0)
    >>> print(f.val, f.der)
    1.0 [0.]
    >>> f = cos(AutoDiff(0.0, N=3))
    >>> print(f.val, f.der)
    1.0 [6.1232340e-17 -1.0000000e+00 -1.8369702e-16]
    """
    half_pi = np.pi / 2
    N = 1
    dk_f = lambda gx, k: np.cos(gx + half_pi * k)  # nth order derivative for cos(x)
    try:
        val_new = np.cos(x.val)
        N = x.N
        if N == 1:
            der_new = -np.sin(x.val) * x.der
        else:
            # N > 1
            der_new = get_n_der_vecs(dk_f, x, N)
    except AttributeError:
        if isinstance(x, float) or isinstance(x, int):
            val_new = np.cos(x)
            # If x is a constant, the derivative of x is 0.
            der_new = np.zeros(1)
        else:
            raise AttributeError('Type error!')
    return AutoDiff(val_new, der_new, N)


def tan_k_order(gx, k):
    """
    Returns the kth order derivative of tan(gx)
    INPUTS
    =======
    gx: Potentially an AutoDiff object
    k: highest derivative order
    RETURNS
    =======
    new AutoDiff object after tan(gx) calculation
    """

    if k == 1:
        return 1 / (np.cos(gx)) ** 2
    else:
        x = AutoDiff(gx, N=k-1)
        f = cos(x)
        dk_f2 = lambda gx2, k2: power_k_order(gx2, -2, k2)
        der_new = get_n_der_vecs(dk_f2, f, k-1)
        return np.float(der_new[k-2])

def tan(x):
    # (tan(x))' = 1/cos(x)^2 * x'
    """
    Returns the value and derivative of a tangent operation: tan(x)
    INPUTS
    =======
    x: an AutoDiff object or a scalar, required, the input variable
    RETURNS
    ========
    an AutoDiff object containing the value and derivative of the expression
    EXAMPLES
    =========
    >>> f = tan(0.0)
    >>> print(f.val, f.der)
    0.0 [0.]
    >>> f = tan(AutoDiff(0.0, N=3))
    >>> print(f.val, f.der)
    0.0 [1. 0. 2.]
    """
    N = 1
    dk_f = lambda gx, k: tan_k_order(gx, k)  # nth order derivative for tan(x)
    try:
        val_new = np.tan(x.val)
        N = x.N
        if N == 1:
            der_new = np.array([1 / (np.cos(x.val)) ** 2 * x.der])
        else:
            # N > 1
            der_new = get_n_der_vecs(dk_f, x, N)
    except AttributeError:
        if isinstance(x, float) or isinstance(x, int):
            val_new = np.tan(x)
            # If x is a constant, the derivative of x is 0.
            der_new = np.zeros(1)
        else:
            raise AttributeError('Type error!')
    return AutoDiff(val_new, der_new, N)


def arcsin_k_order(gx, k):
    """
    Returns the kth order derivative of arcsin(gx)
    INPUTS
    =======
    gx: Potentially an AutoDiff object
    k: highest derivative order
    RETURNS
    =======
    new AutoDiff object after arcsin(gx) calculation
    """
    if k == 1:
        return 1 / np.sqrt(1 - gx**2)
    else:
        x = AutoDiff(gx, N=k-1)
        f = 1 - power(x,2)
        dk_f2 = lambda gx2, k2: power_k_order(gx2, -1/2, k2)
        der_new = get_n_der_vecs(dk_f2, f, k-1)
        return np.float(der_new[k-2])

def arcsin(x):
    # (arcsin(x))' = 1/sqrt(1-(x)^2) * x'
    """
    Returns the value and derivative of an arcsine operation: arcsin(x)
    INPUTS
    =======
    x: an AutoDiff object or a scalar, required, the input variable
    RETURNS
    ========
    an AutoDiff object containing the value and derivative of the expression
    EXAMPLES
    =========
    >>> f = arcsin(0.0)
    >>> print(f.val, f.der)
    0.0 [0.]
    >>> f = arcsin(AutoDiff(0.0, N = 3))
    >>> print(f.val, f.der)
    0.0 [1. 0. 1.]
    """
    N = 1
    dk_f = lambda gx, k: arcsin_k_order(gx, k)  # nth order derivative for arcsin(x)
    if isinstance(x, AutoDiff):
        if x.val < -1 or x.val > 1:
            raise ValueError('Error: Independent variable must be in [-1,1]!')
    try:
        val_new = np.arcsin(x.val)
        N = x.N
        if N == 1:
            der_new = np.array(1 / np.sqrt(1 - x.val ** 2) * x.der)
        else:
            # N > 1
            der_new = get_n_der_vecs(dk_f, x, N)
    except AttributeError:
        if isinstance(x, float) or isinstance(x, int):
            if x < -1 or x > 1:
                raise ValueError('Error: Independent variable must be in [-1,1]!')
            val_new = np.arcsin(x)
            # If x is a constant, the derivative of x is 0.
            der_new = np.zeros(1)
        else:
            raise AttributeError('Type error!')
    return AutoDiff(val_new, der_new, N)


def arccos_k_order(gx, k):
    """
    Returns the kth order derivative of arccos(gx)
    INPUTS
    =======
    gx: Potentially an AutoDiff object
    k: highest derivative order
    RETURNS
    =======
    new AutoDiff object after arccos(gx) calculation
    """
    if k == 1:
        return - 1 / np.sqrt(1 - gx**2)
    else:
        x = AutoDiff(gx, N=k-1)
        f = 1 - power(x,2)
        dk_f2 = lambda gx2, k2: - power_k_order(gx2, -1/2, k2)
        der_new = get_n_der_vecs(dk_f2, f, k-1)
        return np.float(der_new[k-2])

def arccos(x):
    # (arccos(x))' = - 1/sqrt(1-(x)^2) * x'
    """
    Returns the value and derivative of an arccosine operation: arccos(x)
    INPUTS
    =======
    x: an AutoDiff object or a scalar, required, the input variable
    RETURNS
    ========
    an AutoDiff object containing the value and derivative of the expression
    EXAMPLES
    =========
    >>> f = arccos(0.0)
    >>> print(f.val, f.der)
    1.5707963267948966 [0.]
    >>> f = arccos(AutoDiff(0.0, N = 3))
    >>> print(f.val, f.der)
    1.5707963267948966 [-1. 0. -1.]
    """
    N = 1
    dk_f = lambda gx, k: arccos_k_order(gx, k)
    if isinstance(x, AutoDiff):
        if x.val < -1 or x.val > 1:
            raise ValueError('Error: Independent variable must be in [-1,1]!')
    try:
        val_new = np.arccos(x.val)
        N = x.N
        if N == 1:
            der_new = np.array(-1 / np.sqrt(1 - x.val ** 2) * x.der)
        else:
            # N > 1
            der_new = get_n_der_vecs(dk_f, x, N)
    except AttributeError:
        if isinstance(x, float) or isinstance(x, int):
            if x < -1 or x > 1:
                raise ValueError('Error: Independent variable must be in [-1,1]!')
            val_new = np.arccos(x)
            # If x is a constant, the derivative of x is 0.
            der_new = np.zeros(1)
        else:
            raise AttributeError('Type error!')
    return AutoDiff(val_new, der_new, N)


def arctan_k_order(gx, k):
    """
    Returns the kth order derivative of arctan(gx)
    INPUTS
    =======
    gx: Potentially an AutoDiff object
    k: highest derivative order
    RETURNS
    =======
    new AutoDiff object after arctan(gx) calculation
    """
    if k == 1:
        return 1 / (1 + gx ** 2)
    else:
        x = AutoDiff(gx, N = k-1)
        f = 1 + power(x,2)
        dk_f2 = lambda gx2, k2: power_k_order(gx2, -1.0, k2)
        der_new = get_n_der_vecs(dk_f2, f, k-1)
        return np.float(der_new[k-2])

def arctan(x):
    # (arctan(x))' = 1/(1+(x)**2) * x'
    """
    Returns the value and derivative of an arctangent operation: arctan(x)
    INPUTS
    =======
    x: an AutoDiff object or a scalar, required, the input variable
    RETURNS
    ========
    an AutoDiff object containing the value and derivative of the expression
    EXAMPLES
    =========
    >>> f = arctan(0.0)
    >>> print(f.val, f.der)
    0.0 [0.]
    >>> f = arctan(AutoDiff(0.0, N = 3))
    >>> print(f.val, f.der)
    0.0 [1. 0. -2.]
    """
    N = 1
    dk_f = lambda gx, k: arctan_k_order(gx, k)  # nth order derivative for arctan(x)
    try:
        val_new = np.arctan(x.val)
        N = x.N
        if N == 1:
            der_new = np.array(1 / (1 + x.val ** 2) * x.der)
        else:
            # N > 1
            der_new = get_n_der_vecs(dk_f, x, N)
    except AttributeError:
        if isinstance(x, float) or isinstance(x, int):
            val_new = np.arctan(x)
            # If x is a constant, the derivative of x is 0.
            der_new = np.zeros(1)
        else:
            raise AttributeError('Type error!')
    return AutoDiff(val_new, der_new, N)


def sinh_k_order(gx, k):
    """
    Returns the kth order derivative of sinh(gx)
    INPUTS
    =======
    gx: Potentially an AutoDiff object
    k: highest derivative order
    RETURNS
    =======
    new AutoDiff object after sinh(gx) calculation
    """
    if k%2 == 1:
        return np.cosh(gx)
    if k%2 == 0:
        return np.sinh(gx)

def sinh(x):
    # (sinh(x))' = cosh(x) * x'
    """
    Returns the value and derivative of a hyperbolic sine operation: sinh(x)
    INPUTS
    =======
    x: an AutoDiff object or a scalar, required, the input variable
    RETURNS
    ========
    an AutoDiff object containing the value and derivative of the expression
    EXAMPLES
    =========
    >>> f = sinh(0.0)
    >>> print(f.val, f.der)
    0.0 [0.]
    >>> f = sinh(AutoDiff(0.0, N = 3))
    >>> print(f.val, f.der)
    0.0 [1. 0. 1.]
    """
    N = 1
    dk_f = lambda gx, k: sinh_k_order(gx, k)  # nth order derivative for sinh(x)
    try:
        val_new = np.sinh(x.val)
        N = x.N
        if N == 1:
            der_new = np.array(np.cosh(x.val) * x.der)
        else:
            # N > 1
            der_new = get_n_der_vecs(dk_f, x, N)
    except AttributeError:
        if isinstance(x, float) or isinstance(x, int):
            val_new = np.sinh(x)
            # If x is a constant, the derivative of x is 0.
            der_new = np.zeros(1)
        else:
            raise AttributeError('Type error!')
    return AutoDiff(val_new, der_new, N)


def cosh_k_order(gx, k):
    """
    Returns the kth order derivative of cosh(gx)
    INPUTS
    =======
    gx: Potentially an AutoDiff object
    k: highest derivative order
    RETURNS
    =======
    new AutoDiff object after cosh(gx) calculation
    """

    if k%2 == 1:
        return np.sinh(gx)
    if k%2 == 0:
        return np.cosh(gx)


def cosh(x):
    # (cosh(x))' = sinh(x) * x'
    """
    Returns the value and derivative of a hyperbolic cosine operation: cosh(x)
    INPUTS
    =======
    x: an AutoDiff object or a scalar, required, the input variable
    RETURNS
    ========
    an AutoDiff object containing the value and derivative of the expression
    EXAMPLES
    =========
    >>> f = cosh(0.0)
    >>> print(f.val, f.der)
    1.0 [0.]
    >>> f = cosh(AutoDiff(0.0, N = 3))
    >>> print(f.val, f.der)
    1.0 [0. 1. 0.]
    """
    N = 1
    dk_f = lambda gx, k: cosh_k_order(gx, k)  # nth order derivative for cosh(x)
    try:
        val_new = np.cosh(x.val)
        N = x.N
        if N == 1:
            der_new = np.array(np.sinh(x.val) * x.der)
        else:
            # N > 1
            der_new = get_n_der_vecs(dk_f, x, N)
    except AttributeError:
        if isinstance(x, float) or isinstance(x, int):
            val_new = np.cosh(x)
            # If x is a constant, the derivative of x is 0.
            der_new = np.zeros(1)
        else:
            raise AttributeError('Type error!')
    return AutoDiff(val_new, der_new, N)


def tanh(x):
    # (tanh(x))' = (1 - tanh(x)**2) * x'
    """
    Returns the value and derivative of a hyperbolic tangent operation: tanh(x)
    tanh(x)=(e^(2x)-1)/(e^(2x)+1), i.e. f(g(h(x))) with f(x)=(x-1)/(x+1) and g(x)=e^(2x)
    INPUTS
    =======
    x: an AutoDiff object or a scalar, required, the input variable
    RETURNS
    ========
    an AutoDiff object containing the value and derivative of the expression
    EXAMPLES
    =========
    >>> f = tanh(0.0)
    >>> print(f.val, f.der)
    0.0 [0.]
    >>> f = tanh(AutoDiff(0.0, N=3))
    >>> print(f.val, f.der)
    0.0 [1. -0. -2.]
    """
    N = 1
    try:
        val_new = np.tanh(x.val)
        N = x.N
        if N == 1:
            der_new = np.array([(1 - np.tanh(x.val) ** 2) * x.der])
        else:
            # N > 1
            f = lambda x: 1-2/(x+1)
            g = lambda x: exp(2*x)
            der_new = f(g(x)).der
    except AttributeError:
        if isinstance(x, float) or isinstance(x, int):
            val_new = np.tanh(x)
            # If x is a constant, the derivative of x is 0.
            der_new = np.zeros(1)
        else:
            raise AttributeError('Type error!')
    return AutoDiff(val_new, der_new, N)


def logistic(x,x0=0, L=1, k=1):
    """
    Return the value and derivative of a logistic operation L/(1+e^-k(x-x0)}).
    INPUTS
    =======
    x: AutoDiff.variable object or a number.
    x0: float or int, the x value of the sigmoid's midpoint
    L: float or int, curve's maximum value.
    k: float or int, logistic growth rate.
    RETURNS
    ========
    an AutoDiff object containing the value and derivative of the expression
    EXAMPLES
    ========
    >>> f = logistic(0.0)
    >>> print(f.val, f.der)
    0.5 [0.]
    >>> f = logistic(AutoDiff(0.0, N=3))
    >>> print(f.val, f.der)
    0.5 [0.25 0. -0.125]
    """
    N = 1
    try:
        val_new = L / (1 + np.exp(-k * (x.val - x0)))
        der_val = k * (np.exp(-k * (x.val - x0)))
        N = x.N
        if N == 1:
            der_new = np.array([val_new * (val_new / L) * der_val * x.der])
        else:
            # N > 1
            f = lambda x: L * 1/(1+x)
            g = lambda x: exp(-k*(x-x0))
            der_new = f(g(x)).der
    except AttributeError:
        if isinstance(x, float) or isinstance(x, int):
            val_new = L / ( 1 + np.exp(-k * (x-x0)))
            # If x is a constant, the derivative of x is 0.
            der_new = np.zeros(1)
        else:
            raise AttributeError('Type error!')
    return AutoDiff(val_new, der_new, N)



import setuptools


with open("README.md", "r", encoding="utf-8") as fh:
    README = fh.read()

with open("requirements.txt", "r", encoding="utf-8") as f:
    REQUIREMENTS = f.readlines()

PKGS = ["SmartDiff"]

setuptools.setup(
    name="Smartdiff",
    version="0.0.15",
    author="Harvard SmartDiff Group",
    author_email="harvard.smartdiff@yahoo.com",
    description="A smart package for Automatic Differentiation",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/SmartDiff/cs107-FinalProject",
    packages=PKGS,
    install_requires=REQUIREMENTS,
    include_package_data=True,
    package_data={'SmartDiff': ['SmartDiff/GUI/*.ui']},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import sys
from PyQt5 import QtGui, QtCore, QtWidgets, uic
import math
from sympy import symbols
from SmartDiff.preprocess.pyexpr_formatter import PyExpression_Formatter
from SmartDiff.solvers.element_op import *
from SmartDiff.solvers.multivariate import *

# global Ui_MainWindow, Ui_SecondDiag
Ui_MainWindow, QtBaseClass = uic.loadUiType('GUI/step1.ui')  # .ui drawn in Qt Designer
Ui_FourthDiag, QtBaseClass4 = uic.loadUiType('GUI/step4.ui')  # .ui drawn in Qt Designer
Ui_FifthDiag, QtBaseClass5 = uic.loadUiType('GUI/step5.ui')  # .ui drawn in Qt Designer

class MainWindow(QtWidgets.QMainWindow, Ui_MainWindow):

    def __init__(self):
        QtWidgets.QMainWindow.__init__(self)
        Ui_MainWindow.__init__(self)
        self.setWindowTitle("SmartDiff")
        self.setupUi(self)
        self.Instructions.setWordWrap(True)

        self.FuncDim = 1  # default
        self.InputDim = 1  # default

        self.order = self.setupOrder()
        self.val = np.zeros(1)
        self.func = None

        # once OK button is pressed, go to step 2 and 3 and 4
        self.OKButton.clicked.connect(self.onClickOK)

    def setupOrder(self):
        '''
        :param: None
        :return:
        int, user input: the order of derivative to calculate
        '''
        order, okPressed = QtWidgets.QInputDialog.getInt(self, "Step 0: Input the order of derivative", "N = ",
                                                         value=1, min=1)
        if okPressed:
            return order

    def SetDimInput(self):
        '''
        Set up the function and input dimensions
        :return:
        user input of function dimension and input dimension
        '''
        self.FuncDim = int(self.FuncDimBox.currentText())
        self.InputDim = int(self.InputDimBox.currentText())

    def DisplayOrderWarning(self):
        '''
        Display warnings if the order is too high
        Reset the self.FuncDim and self.InputDim based on self.order
        :return:
        None
        '''
        warning_msg = ""
        if self.order > 10:
            self.FuncDim = 1
            self.InputDim = 1
            warning_msg += "Warning: the order of derivative is high, so the computation may take a long time. "
            if self.FuncDim > 1:
                self.FuncDim = 1
                warning_msg += "Reset function dimension to 1. "
            if self.InputDim > 1:
                self.InputDim = 1
                warning_msg += "Reset input dimension to 1. "
        elif self.order > 2:
            if self.FuncDim > 1:
                self.FuncDim = 1
                warning_msg += "Warning: Reset function dimension to 1. "
            if self.InputDim > 1:
                self.InputDim = 1
                warning_msg += "Warning: Reset input dimension to 1. "
        elif self.order == 2:
            if self.FuncDim > 1:
                self.FuncDim = 1
                warning_msg += "Warning: Reset function dimension to 1. Hessian is not supported for vector-valued " \
                               "functions."
        self.ErrMsg.setText(warning_msg)
        self.ErrMsg.setWordWrap(True)

    def PointEval(self):
        '''
        get user input values for each variable
        :return:
        np array of the user input (size = self.InputDim from step 1)
        '''
        xs = symbols('x:%d' % self.InputDim)
        x_list = []
        for i in range(self.InputDim):
            x_list.append(self._PointEval(str(xs[i])))
        return x_list

    def _PointEval(self, string):
        '''
        :param string: x0, x1, x2, ... to put in the QInputDialog box title
        :return:
        double, user input, default 0, min -100, max 100, up to 4 decimals
        '''
        # Need to make the dialog window larger to show the title
        num, okPressed = QtWidgets.QInputDialog.getDouble(self, "Step 2: Input the evaluating point", string+" = ",
                                                          0, -100, 100, 4)
        if okPressed and num != '':
            return num

    def FuncEval(self):
        '''
        get user input function to differentiate
        :return:
        a list of user input function (each element is a component of the function)
        '''
        fs = symbols('f:%d' % self.FuncDim)
        f_list = []
        for i, f in enumerate(fs):
            f_list.append(self._FuncEval(str(f)))
        return f_list

    def _FuncEval(self, string):
        '''
        :param string: f0, f1, f2,... to put in the QInputDialog box title
        :return:
        str, user input
        '''
        # Need to make the dialog window larger to show the title
        func, okPressed = QtWidgets.QInputDialog.getText(self, "Step 3: Input the function", string+" = ",
                                                         QtWidgets.QLineEdit.Normal, "")
        if okPressed and func != '':
            return func

    def onClickOK(self):
        '''
        Trigger step 2: User puts in values of the variables to evaluate
        Trigger step 3: User puts in the function to evaluate and differentiate (working on this now)
        Trigger step 4: User confirms the input is correct and chooses whether to show the function value
        :return: None
        '''
        self.SetDimInput()
        self.DisplayOrderWarning()
        # step 2
        self.val = self.PointEval()
        # step 3
        self.func = self.FuncEval()
        # step 4
        dlg4 = FourthDiag(self.InputDim, self.FuncDim, self.val, self.func, self.order)
        dlg4.exec_()


class FourthDiag(QtWidgets.QDialog, Ui_FourthDiag):

    def __init__(self, InputDim, FuncDim, val, func, order):
        QtWidgets.QDialog.__init__(self)
        # load a dialogue based on user input from step one
        self.InputDim = InputDim
        self.FuncDim = FuncDim
        self.val = val
        self.func = func
        self.order = order

        self.setupUi(self)
        Ui_FourthDiag, QtBaseClass4 = uic.loadUiType('GUI/step4.ui')
        Ui_FourthDiag.__init__(self)
        self.DisVal = False
        # populate the boxes based on user input in step 2 and 3
        self.SetupValFunc()
        # checkBox to select whether the user wants to show the function values
        self.checkBox.clicked.connect(self.setDisVal)
        # click OK button to start the computation
        self.OKButton.clicked.connect(self.onClickOK)

    def SetupValFunc(self):
        '''
        Display the function expressions, variable value input based on user input
        :return: None
        '''
        xs = symbols('x:%d' % self.InputDim)
        fs = symbols('f:%d' % self.FuncDim)
        var_msg = ""
        func_msg = ""
        for i, x in enumerate(self.val):
            var_msg += f"{str(xs[i])}={self.val[i]}\n"
        for i, f in enumerate(self.func):
            func_msg += f"{str(fs[i])}={self.func[i]}\n"
        self.ValInput.setPlainText(var_msg)
        self.ValInput.setReadOnly(True)
        self.ValInput.setStyleSheet("background: white; color: black")
        self.FuncInput.setPlainText(func_msg)
        self.FuncInput.setReadOnly(True)
        self.FuncInput.setStyleSheet("background: white; color: black")

    def setDisVal(self):
        '''
        Set whether the GUI displays the function value in addition to the derivative
        :return:
        None
        '''
        self.DisVal = not self.DisVal

    def onClickOK(self):
        '''
        sets up step 5
        :return:
        The function value and derivative at the specific point from user input
        '''
        # compute function value and derivative and get error messsage

        val, der, msg = self.compValDer()
        # step 5
        dlg5 = FifthDiag(val, der, msg, self.order, self.DisVal)
        dlg5.exec_()

    def compValDer(self):
        '''
        Format the user input in step 2 and 3 into strings and call AD modules in solvers
        :return:
        np.array, np.array, str: function value, function derivative, error message
        '''
        # instantiate a formatter object
        formatter = PyExpression_Formatter()
        # xs = symbols('x:%d' % self.InputDim)
        # var_map = {str(xs[i]): self.val[i] for i in range(self.InputDim)}
        # var_map.update({"x": self.val[0]})
        # if self.InputDim == 2:
        #     var_map.update({"y": self.val[1]})
        # if self.InputDim == 3:
        #     var_map.update({"y": self.val[1], "z": self.val[2]})
        func_map = {"pi": math.pi, "e": math.e, "power": power, "log": log, "exp": exp, "sqrt": sqrt, "sin": sin,
                    "cos": cos, "tan": tan, "arcsin": arcsin, "arccos": arccos, "arctan": arctan, "sinh": sinh,
                    "cosh": cosh, "tanh": tanh}
        # var_map.update(func_map)

        err_msg = ""
        # Get user input and check if it's valid
        for i, f in enumerate(self.func):
            is_valid = formatter.is_valid_input(f)
            if is_valid == 1:
                return np.zeros(1), np.zeros(1), f"Input function has unmatched parenthesis in f{i}!"
            elif is_valid != 0:
                return np.zeros(1), np.zeros(1), f"Input function contains invalid character in f{i}!"

        # proceed if all the functions pass user input checking
        try:
            if self.FuncDim == 1:
                if self.order == 1:
                    val, der = get_val_jacobian(self.func, self.val, eval_func_map=func_map)
                    return val, der, err_msg
                elif self.order == 2:
                    val, _ = get_val_jacobian(self.func, self.val, eval_func_map=func_map)
                    der = get_hessian(self.func[0], self.val, eval_func_map=func_map)
                    return val, der, err_msg
                else:  # self.order > 2 and self.FuncDim == 1 and self.InputDim == 1
                    var_map = {'x': AutoDiff(self.val[0], N=self.order), 'x1': AutoDiff(self.val[0], N=self.order), 'x0': AutoDiff(self.val[0], N=self.order)}
                    var_map.update(func_map)
                    AD_out = eval(self.func[0], var_map)
                    try:
                        val = AD_out.val
                        der = AD_out.der[-1]
                    except AttributeError:
                        val = AD_out
                        der = 1
                    return val, der, err_msg
            else:
                # self.FuncDim > 1 and self.order == 1
                val, der = get_val_jacobian(self.func, self.val, eval_func_map=func_map)
                return val, der, err_msg
        except ValueError as e:
            return np.zeros(1), np.zeros(1), str(e)+" "
        except AttributeError as e:
            return np.zeros(1), np.zeros(1), str(e)+" "
        except ZeroDivisionError as e:
            return np.zeros(1), np.zeros(1), str(e)+" "


class FifthDiag(QtWidgets.QDialog, Ui_FifthDiag):

    def __init__(self, Val, Der, Msg, Order, DisVal):
        QtWidgets.QDialog.__init__(self)
        # load a dialogue based on user input from step one
        self.val = Val
        self.der = Der
        self.msg = Msg
        self.order = Order
        self.DisVal = DisVal
        Ui_FifthDiag, QtBaseClass5 = uic.loadUiType('GUI/step5.ui')
        Ui_FifthDiag.__init__(self)
        self.setupUi(self)

        # populate the boxes based on user input in step 2 and 3
        self.ResultDisplay()
        # stop the program when user clicks quit
        self.quitButton.clicked.connect(self.onClickQuit)

    def ResultDisplay(self):
        '''
        Display the results of the SmartDiff, based on user input
        :return:
        None
        '''
        self.ErrMsg.setWordWrap(True)
        self.FuncVal.setReadOnly(True)
        self.FuncVal.setStyleSheet("background: white; color: black")
        self.DerVal.setReadOnly(True)
        self.DerVal.setStyleSheet("background: white; color: black")

        if self.msg == "":
            self.ErrMsg.setText("Success! See results below")
            if self.DisVal:
                self.FuncVal.setPlainText(str(self.val))
            else:
                self.FuncVal.setPlainText("N/A")
            der_msg = f"{self.order}-order derivative:\n"
            self.DerVal.setPlainText(der_msg + str(self.der))
        else:
            self.FuncVal.setPlainText("N/A")
            self.DerVal.setPlainText("N/A")
            self.ErrMsg.setText("Failure: " + self.msg +
                                "Close windows of step 4 and 5 and start again from step 1.")

    def onClickQuit(self):
        '''
        Stop the program when user clicks quit
        :return:
        None
        '''
        sys.exit()


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    window = MainWindow()
    window.show()
    app.exec_()
